/*************************************************************************/
/*************************************************************************/
/* H. Timucin OZDEMIR                                                    */
/* Date : February, 1997                                                 */
/* File Name : FuzzyInfEngine.cc                                         */
/*************************************************************************/
/*************************************************************************/

#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include "FuzzyInfEngine.h"
#include "RuleMatrix.h"
#include "Defuzzifier.h"
#include "Vector.h"
#include "FuzzyVariable.h"
#include "FunctionBag.h"
#include "../deffunc.h"

FuzzyInfEngine::FuzzyInfEngine(int _in,int _out)
{ 
   NumberOfInput = _in;
   NumberOfOutput = _out;
   myRules = new RuleMatrix(_in,_out);
   inputVarIndx = outputVarIndx = 0;
   inputFuzzyVars = outputFuzzyVars = NULL;

}// end of constructor

FuzzyInfEngine::~FuzzyInfEngine()
{
   //if(inputVarIndx) delete [] inputFuzzyVars;
   //if(outputVarIndx) delete [] outputFuzzyVars;
   //delete myRules;

   cout << " Deallocate FuzzyInfEngine " << endl;

}// end of destructor

void FuzzyInfEngine::addInput(FuzzyVariable *inputVar)
{  FuzzyVariable **Temp = new FuzzyVariable*[inputVarIndx+1];
   int i;

   for(i=0;i<inputVarIndx;i++) Temp[i]=inputFuzzyVars[i];
   Temp[i] = inputVar;
   if(inputVarIndx) delete [] inputFuzzyVars;
   inputVarIndx ++;
   inputFuzzyVars = Temp;

}// end of addInput()

void FuzzyInfEngine::addOutput(FuzzyVariable *inputVar)
{  FuzzyVariable **Temp = new FuzzyVariable*[outputVarIndx+1];
   int i;

   for(i=0;i<outputVarIndx;i++) Temp[i]=outputFuzzyVars[i];
   Temp[i] = inputVar;
   if(outputVarIndx) delete [] outputFuzzyVars;
   outputVarIndx ++;
   outputFuzzyVars = Temp;

}// end of addOutput()

// Output 0 means that everything is OK.

FuzzyInfEngine::addRule(Vector inDef,Vector outDef)
{
   return (*myRules).addRule(inDef,outDef);

}// end of addRule()


Vector FuzzyInfEngine::RunEngine(Vector input, Defuzzifier *deff)
{
   FunctionBag * ruleOut = new FunctionBag[NumberOfOutput];
   float *outval = new float[NumberOfOutput];
   int i,j,k;
   int numberOfRules;
   float cutOff;
   Vector inputPattern;
   Vector outputPattern;
   Function *outFuzzy;


   for(i=0;i<NumberOfOutput;i++) outval[i] = 0.0;

   if(input.size() == NumberOfInput)
   {
      numberOfRules = (*myRules).size();

      cout << " We are going to look " << numberOfRules << " num rules " << endl;
      for(i=0;i<numberOfRules;i++)
      {  inputPattern = (*myRules).getInputVector(i);

         //cout << "Rule #: "<<i<<" I pattern :"<<inputPattern<< endl;

         cutOff = 1.0;

         for(j=0;j<NumberOfInput;j++)
         {  k = inputPattern.getElement(j); // class
	    if(k)
	     cutOff = minimum(cutOff, 
		(*inputFuzzyVars[j]).getValueWithClassIndex(input.getElement(j),k-1));
         }// end of for

	 //cout << " cutOff = " << cutOff << endl;

	 outputPattern = (*myRules).getOutputVector(i);

         //cout << "Rule #: "<<i<<" O pattern :"<<outputPattern<< endl;

	 for(j=0;j<NumberOfOutput;j++)
	 {  k = outputPattern.getElement(j);
	    if(k)
	    {  outFuzzy = (*outputFuzzyVars[j]).copyWithCutOffClassIndex(cutOff,k-1);
	       ruleOut[j].addFunction(outFuzzy);

	       //cout << *outFuzzy << " is added to output " << j << endl;
	    }

	 }// end of for


      }// end of for

      //cout << " Before defuzz " << endl;

      //for(j=0;j<NumberOfOutput;j++) 
      //    cout << "result # " << j << " " << ruleOut[j];

      for(j=0;j<NumberOfOutput;j++) 
	  outval[j] = (*deff).defuzzy(&ruleOut[j]);

      for(j=0;j<ruleOut[0].getNumberOfFunctions();j++) 
      {outFuzzy = ruleOut[0].getFunctionWithIndx(j);
      if(outFuzzy)
      {  char fileName[12];  
	 sprintf(fileName,"First%d.out",j);
	 ofstream ost(fileName);
	 for(float k=0;k<1.0;k+=0.01)
		   ost << k<<" "<< (*outFuzzy).getValue(k) << endl;
         ost.close();
      }// end of if
      }

      ofstream ost("Var0.out");
      for(float k=0;k<1.0;k+=0.01)
	  ost << k<<" "<< ruleOut[0].getValue(k) << endl;
      ost.close();
      

   }// end of if

   Vector ret(outval,NumberOfOutput);

   return ret;

}// end of RunEngine()



ostream& operator<<(ostream& os,const FuzzyInfEngine& RHS)
{  os << " Fuzzy Inference Engine with " << RHS.NumberOfInput ;
   os << " input " <<  RHS.NumberOfOutput << " output " << endl;
   os << " Rule Matrix is " << endl << *RHS.myRules << endl;
   return os;
}// end of operator<<()


